-- @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
-- @@@@@@@@@@@@  ASSETS->TBL_ASSETS_MASTER TBL_ASSETS_MASTER TBL_ASSETS_MASTER  @@@@@@@@@@@@
-- @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
------------------------------------------------------------------------------------------
            CREATE OR REPLACE FUNCTION fn_get_all_assets_master() RETURNS SETOF assets.tbl_assets_master AS $$
                DECLARE

                BEGIN
                    RETURN QUERY SELECT * FROM assets.tbl_assets_master WHERE int_sys_action_id != -1;  
                END;
            $$ LANGUAGE plpgsql;
        

------------------------------------------------------------------------------------------
            CREATE OR REPLACE FUNCTION fn_get_all_assets_master_with_action_user() 
                            RETURNS TABLE (pk_assets_master_id BIGINT,

								fk_account_id BIGINT,
								fk_ad_cr_account_id BIGINT,
								fk_depreciation_dr_account_id BIGINT,
								sin_depreciation_based_on SMALLINT,
								dbl_depreciation_perc DOUBLE PRECISION,

								sin_in_code_set_branch_code SMALLINT,
								vhr_code_prefix VARCHAR,
								bin_code_part_last_number BIGINT) AS $$
                DECLARE

                BEGIN
                    RETURN QUERY SELECT mt.pk_assets_master_id,

					mt.fk_account_id,
					mt.fk_ad_cr_account_id,
					mt.fk_depreciation_dr_account_id,
					mt.sin_depreciation_based_on,
					mt.dbl_depreciation_perc,

					mt.sin_in_code_set_branch_code,
					mt.vhr_code_prefix,
					mt.bin_code_part_last_number
                        FROM assets.tbl_assets_master mt 
                            LEFT JOIN authentication.tbl_user u1
                                ON mt.fk_created_user_id =  u1.pk_user_id
                            LEFT JOIN authentication.tbl_user u2
                                ON mt.fk_last_modified_user_id =  u2.pk_user_id
                            LEFT JOIN authentication.tbl_user u3
                                ON mt.fk_deleted_user_id =  u3.pk_user_id
                        WHERE mt.int_sys_action_id != -1  
                        ORDER BY ;
                END;
            $$ LANGUAGE plpgsql;
        

------------------------------------------------------------------------------------------

            CREATE OR REPLACE PROCEDURE sp_insert_or_update_assets_master(jsnAssetsMaster JSONB, jsnSqlRes OUT JSON) AS $$
                DECLARE
                    objSpOutRes TYP_SP_OUT_RES;
                    objFnCheckRes TYP_FN_CHECK_RES;
                    dtmCurrent TIMESTAMP WITH TIME ZONE;
                BEGIN
                    SELECT NOW() INTO dtmCurrent;

                    -- // Duplication Checking in DB FUCTION(Insert and Update)
                    SELECT * FROM  fn_check_data_uniqueness('assets.tbl_assets_master'::VARCHAR, 'sin_in_code_set_branch_code'::VARCHAR, jsnAssetsMaster->>'sinInCodeSetBranchCode'::VARCHAR, 'pk_assets_master_id'::VARCHAR, CAST(jsnAssetsMaster->>'binAssetsMasterId' AS BIGINT)::BIGINT) INTO objFnCheckRes;
                    IF objFnCheckRes.sin_sql_check_status = -1::SMALLINT THEN
                        objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                        objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                        objSpOutRes.strSqlCustomMessage := objFnCheckRes.vhr_sql_check_message;

                        objSpOutRes.strSqlCustomErrorType := 'DATABASE_VALIDATION_ERROR';

                        SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                        RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                        RETURN;
                    END IF;

                    IF CAST(jsnAssetsMaster->>'binAssetsMasterId' AS BIGINT)::BIGINT = 0::BIGINT THEN
                        -- // Count Exced or other checking License related in DB FUCTION(Insert)
                        

                        --jsnAssetsMaster.intSysActionId = 0;
                        --jsnAssetsMaster.binCreatedUserId = jsnAssetsMaster.linLoginUserId;
                        --jsnAssetsMaster.dtmCreated = jsnAssetsMaster.dtmCurrentDateTime; --SELECT NOW();

                        INSERT INTO assets.tbl_assets_master (fk_account_id,
						fk_ad_cr_account_id,
						fk_depreciation_dr_account_id,
						sin_depreciation_based_on,
						dbl_depreciation_perc,

						sin_in_code_set_branch_code,
						vhr_code_prefix,
						bin_code_part_last_number)
                            VALUES(CAST(jsnAssetsMaster->>'binAccountId' AS BIGINT),
						CAST(jsnAssetsMaster->>'binAdCrAccountId' AS BIGINT),
						CAST(jsnAssetsMaster->>'binDepreciationDrAccountId' AS BIGINT),
						CAST(jsnAssetsMaster->>'sinDepreciationBasedOn' AS SMALLINT),
						CAST(jsnAssetsMaster->>'dblDepreciationPerc' AS DOUBLE PRECISION),

						CAST(jsnAssetsMaster->>'sinInCodeSetBranchCode' AS SMALLINT),
						jsnAssetsMaster->>'strCodePrefix',
						CAST(jsnAssetsMaster->>'binCodePartLastNumber' AS BIGINT)) RETURNING pk_assets_master_id INTO objSpOutRes.binSqlPk;
                    ELSE
                        -- // UPDATE Check already update another usersin DB FUCTION(Update and Delete)
                        SELECT * FROM  fn_check_already_update_or_delete_another_user('assets.tbl_assets_master'::VARCHAR, 'pk_assets_master_id'::VARCHAR, CAST(jsnAssetsMaster->>'binAssetsMasterId' AS BIGINT)::BIGINT, CAST(jsnAssetsMaster->>'intSysActionId' AS INT)::INT) INTO objFnCheckRes;
                        IF objFnCheckRes.sin_sql_check_status = -1::SMALLINT THEN
                            objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                            objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                            objSpOutRes.strSqlCustomMessage := objFnCheckRes.vhr_sql_check_message;

                            objSpOutRes.strSqlCustomErrorType := 'DATABASE_VALIDATION_ERROR';

                            SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                            RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                            RETURN;
                        END IF;

                        --jsnAssetsMaster.intSysActionId = jsnAssetsMaster.intSysActionId + 1;
                        --jsnAssetsMaster.binLastModifiedUserId = jsnAssetsMaster.linLoginUserId;
                        --jsnAssetsMaster.dtmLastModified = jsnAssetsMaster.dtmCurrentDateTime; --SELECT NOW();

                        UPDATE assets.tbl_assets_master SET fk_account_id = CAST(jsnAssetsMaster->>'binAccountId' AS BIGINT),
						fk_ad_cr_account_id = CAST(jsnAssetsMaster->>'binAdCrAccountId' AS BIGINT),
						fk_depreciation_dr_account_id = CAST(jsnAssetsMaster->>'binDepreciationDrAccountId' AS BIGINT),
						sin_depreciation_based_on = CAST(jsnAssetsMaster->>'sinDepreciationBasedOn' AS SMALLINT),
						dbl_depreciation_perc = CAST(jsnAssetsMaster->>'dblDepreciationPerc' AS DOUBLE PRECISION),

						sin_in_code_set_branch_code = CAST(jsnAssetsMaster->>'sinInCodeSetBranchCode' AS SMALLINT),
						vhr_code_prefix = jsnAssetsMaster->>'strCodePrefix',
						bin_code_part_last_number = CAST(jsnAssetsMaster->>'binCodePartLastNumber' AS BIGINT) 
                        WHERE pk_assets_master_id = CAST(jsnAssetsMaster->>'binAssetsMasterId' AS BIGINT);
                    END IF;

                    --COMMIT;
                    objSpOutRes.sinSqlCustomStatus := 1::SMALLINT;
                    objSpOutRes.strSqlCustomStatusCode := 'SUCCESS';
                    objSpOutRes.strSqlCustomMessage := 'AssetsMaster Saved';
                    objSpOutRes.dtmCurrent = dtmCurrent;
                    
                    SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                    RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                    RETURN;
                EXCEPTION WHEN OTHERS THEN
                    ROLLBACK;
                    objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                    objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                    objSpOutRes.strSqlCustomMessage := 'AssetsMaster Save Failed';

                    objSpOutRes.strSqlCustomErrorType := 'DATABASE_SYSTEM_ERROR';

                    GET STACKED DIAGNOSTICS 
				        objSpOutRes.strSqlSysMessage := MESSAGE_TEXT,
                    	objSpOutRes.strSqlSysExceptionDetail := PG_EXCEPTION_DETAIL,
                    	objSpOutRes.strSqlSysExceptionHint := PG_EXCEPTION_CONTEXT;
                
                    SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                    RAISE EXCEPTION 'jsnSqlRes: %', jsnSqlRes;
                    RETURN;
                END;
            $$ LANGUAGE plpgsql;
        

------------------------------------------------------------------------------------------

            CREATE OR REPLACE PROCEDURE sp_delete_assets_master(binAssetsMasterId BIGINT, intSysActionId INT, binDeletedUserId BIGINT, jsnSqlRes OUT JSON) AS $$
                DECLARE
                    objSpOutRes TYP_SP_OUT_RES;
                    objFnCheckRes TYP_FN_CHECK_RES;
                    dtmCurrent TIMESTAMP WITH TIME ZONE;
                BEGIN
                    SELECT NOW() INTO dtmCurrent;

                    -- // FOREIGN KEY REFERENCES Checking in DB FUCTION(Delete)
                    SELECT * FROM  fn_check_foreignkey_referance('tbl_assets_master'::VARCHAR, 'pk_assets_master_id'::VARCHAR, binAssetsMasterId::BIGINT, '{}'::VARCHAR ARRAY) INTO objFnCheckRes;
                    IF objFnCheckRes.sin_sql_check_status = -1::SMALLINT THEN
                        objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                        objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                        objSpOutRes.strSqlCustomMessage := objFnCheckRes.vhr_sql_check_message;

                        objSpOutRes.strSqlCustomErrorType := 'DATABASE_VALIDATION_ERROR';

                        SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                        RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                        RETURN;
                    END IF;

                    -- // UPDATE Check already update another users in DB FUCTION(Update and Delete)
                    SELECT * FROM  fn_check_already_update_or_delete_another_user('assets.tbl_assets_master'::VARCHAR, 'pk_assets_master_id'::VARCHAR, CAST(binAssetsMasterId AS BIGINT)::BIGINT, intSysActionId::INT) INTO objFnCheckRes;
                    IF objFnCheckRes.sin_sql_check_status = -1::SMALLINT THEN
                        objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                        objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                        objSpOutRes.strSqlCustomMessage := objFnCheckRes.vhr_sql_check_message;

                        objSpOutRes.strSqlCustomErrorType := 'DATABASE_VALIDATION_ERROR';

                        SELECT * FROM  fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                        RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                        RETURN;
                    END IF;

                    --dtmDeleted = SELECT NOW();
                    UPDATE assets.tbl_assets_master SET int_sys_action_id = -1::INT,
                                fk_deleted_user_id = binDeletedUserId,
                                dtm_deleted = dtmCurrent
                        WHERE pk_assets_master_id = binAssetsMasterId;

                    --COMMIT;
                    objSpOutRes.sinSqlCustomStatus := 1::SMALLINT;
                    objSpOutRes.strSqlCustomStatusCode := 'SUCCESS';
                    objSpOutRes.strSqlCustomMessage := 'AssetsMaster Deleted';
                    objSpOutRes.dtmCurrent = dtmCurrent;

                    SELECT * FROM fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                    RAISE NOTICE 'jsnSqlRes: %', jsnSqlRes;
                    RETURN;
                EXCEPTION WHEN OTHERS THEN
                    ROLLBACK;
                    objSpOutRes.sinSqlCustomStatus := -1::SMALLINT;
                    objSpOutRes.strSqlCustomStatusCode := 'ERROR';
                    objSpOutRes.strSqlCustomMessage := 'AssetsMaster Delete Failed';

                    objSpOutRes.strSqlCustomErrorType := 'DATABASE_SYSTEM_ERROR';

                    GET STACKED DIAGNOSTICS 
				        objSpOutRes.strSqlSysMessage := MESSAGE_TEXT,
                    	objSpOutRes.strSqlSysExceptionDetail := PG_EXCEPTION_DETAIL,
                    	objSpOutRes.strSqlSysExceptionHint := PG_EXCEPTION_CONTEXT;

                    SELECT * FROM fn_convert_type_out_object_to_out_json(objSpOutRes) INTO jsnSqlRes;
                    RAISE EXCEPTION 'jsnSqlRes: %', jsnSqlRes;
                    RETURN;
                END;
            $$ LANGUAGE plpgsql;
        

